#include <memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "comm.h"

#ifndef LIFECALC
#include "lifecalc.h"
#endif

extern bmsize thesize;

int		BfreeMemory = 0;  	/* wird von Init u.a. benutzt */
BufferType	*BFieldBaseAddress;     /* Basisadresse des Zellenfeldes */
BufferType	*BFieldBufferAddress;	/* Pufferadresse zwecks Berechnung */
BufferType	*BFieldAddress;		/* Adresse des Arbeitsfeldes, d.h.
					der 2. Zeile */
BufferType	*BFieldCurrentAddress;
int		BFieldWidth = 0;	/* Breite des Zellenfeldes in Bytes */
int		BFieldHeight = 0;	/* Hoehe d.Zf. in Zeilen */
int		BFieldSize = 0;		/* Gesamtgroesse des Zellfeldes in Bytes */
int		BSlaveID = 0;
int		BDebugID = -1;		/* ID des Prozesses, der Debugs ausspucken
					soll */

int		BRuleBirthMin	= 3;
int		BRuleBirthMax	= 3;
int		BRuleSurviveMin	= 2;
int		BRuleSurviveMax	= 3;	/* eigentlich 2, aber aequivalent und schneller */

#define BsSetCell(f,x,y)	\
	(*((f) + (y) * thesize.lofl + ((x)>>3)) |= (0x01 << ((x) & 0x07)))

#define BsResetCell(f,x,y)	\
	(*((f) + (y) * thesize.lofl + ((x)>>3)) &= ~(0x01 << ((x) & 0x07)))

#define BsGetCell(f,x,y)	\
	(!! (*((f) + (y) * thesize.lofl + ((x)>>3)) & (0x01 << ((x) & 0x07))))

int BsGetNeighbours (BufferType *Field, int xCell, int yCell)
{
    int	neighbours;

    neighbours	    = BsGetCell(Field,xCell-1,yCell-1);
    neighbours     += BsGetCell(Field,xCell,yCell-1);
    neighbours     += BsGetCell(Field,xCell+1,yCell-1);
    neighbours     += BsGetCell(Field,xCell-1,yCell+1);
    neighbours     += BsGetCell(Field,xCell,yCell+1);
    neighbours     += BsGetCell(Field,xCell+1,yCell+1);
    neighbours     += BsGetCell(Field,xCell-1,yCell);
    neighbours     += BsGetCell(Field,xCell+1,yCell);

    return(neighbours);
}

#define BsClearField(f)		\
	memset(f, '\0', thesize.length + 2*thesize.lofl)

#define BsCopyField(s,d)       	\
	memcpy(d, s, thesize.length + 2*thesize.lofl)


int	BsNewGeneration(void)
{
/* Hier wird die Berechnung der neuen Generation ausgefuehrt.
Rueckgabe: -1 ==> alle Zellen tot
	    0 ==> gemischt
	    1 ==> alle Zellen leben */

    int	xCell,yCell;
    int	xBorder,yBorder;
    int	neighbours;
    int	empty,full;	/* alle Zellen tot/lebendig ? */
    BufferType *Zeile1,*Zeile2,*Zeile3,*SetZeile;
    int CurrentRow,CurrCell;
    int NextRow,NextCell;
    int PrevRow;
    BufferType Bitmask,SetBitmask;
    
    xBorder = thesize.xsize - 1;
    yBorder = thesize.ysize + 1;

    empty	= 1;
    full	= 1;

	  Zeile1 = BFieldBaseAddress;
      	  Zeile2 = BFieldBaseAddress+thesize.lofl;
      	  Zeile3 = Zeile2+thesize.lofl;
    	  SetZeile = BFieldBufferAddress+thesize.lofl;

    BsClearField(BFieldBufferAddress);
    for(yCell = 1;yCell < yBorder;yCell++)
      {
      /* Hier wird die Zellmatrix zeilenweise durchlaufen */
/* Allgemeines zur Berechnung der Anzahl an Nachbarzellen:
um nicht bei jedem Zellzugriff eine neue Adressberechnung durchfuehren
zu muessen, wird zu Beginn einer jeden Zeile je ein Zeiger auf das erste
Byte der Zeile und auf die naechsthoehere bzw. -niedere gesetzt.
Im Verlauf der Berechnung werden jetzt nur noch die Zeiger ggfs. neu
gesetzt und innerhalb des jeweiligen Bytes wird mit Masken gearbeitet. */	
      	  Bitmask = 0x02;
      	  SetBitmask = 0x01;
	  PrevRow = 0;
	  CurrentRow = 0;
	  if(Bitmask & *Zeile1)	CurrentRow++;
	  if(Bitmask & *Zeile3) CurrentRow++;
	  if(Bitmask & *Zeile2)	CurrCell = 1;
	  else			CurrCell = 0;
	  Bitmask <<= 1;
	  SetBitmask <<= 1;
	  for(xCell = 1;xCell < xBorder;xCell++)
	    {
	    /* Hier wird sie Spaltenweise durchlaufen */	
	    	NextRow = 0;
	    	if(Bitmask & *Zeile1)	NextRow++;
	    	if(Bitmask & *Zeile3)	NextRow++;
		if(Bitmask & *Zeile2)	NextCell = 1;
		else			NextCell = 0;

		neighbours = PrevRow+CurrentRow+NextRow+NextCell;
		if((neighbours<=BRuleSurviveMax)&&(neighbours>=BRuleSurviveMin)&&CurrCell)
			*SetZeile |= SetBitmask;
		if((neighbours<=BRuleBirthMax)&&(neighbours>=BRuleBirthMin))
			*SetZeile |= SetBitmask;

		PrevRow = CurrentRow+CurrCell;
		CurrentRow = NextRow;
		CurrCell = NextCell;

		Bitmask <<= 1;
		if(!Bitmask)
		{
			Bitmask = 0x01;
			Zeile1++;
			Zeile2++;
			Zeile3++;
		}
		SetBitmask <<= 1;
		if(!SetBitmask)
		{
			SetZeile++;
			SetBitmask = 0x01;
		}

	    }/* of x -> */
	if(Bitmask != 0x01)
	{
		Zeile1++;
		Zeile2++;
		Zeile3++;
	}
	if(SetBitmask != 0x01)
	{
		SetZeile++;
	}
    } /* of y -> */

 /* BsCopyField(BFieldBufferAddress,BFieldBaseAddress);*/
    BFieldBaseAddress = BFieldBufferAddress;
    BFieldBufferAddress = BFieldCurrentAddress;
    BFieldCurrentAddress = BFieldBaseAddress;
    if(empty == 1)
      return(-1);
    else if(full == 1)
      return(1);
    else
      return(0);
}


int	BSet (void)
{
    /* Soweit alles ok, nun Speicher allokieren und beschreiben */
    BFieldWidth  = thesize.lofl;
    BFieldHeight = thesize.ysize + 2;
    BFieldSize = BFieldWidth * BFieldHeight;

    if((DEBUG == ON) && (BDebugID == BSlaveID))
      {
	  fprintf(stderr,"\nPID[%3d] Feldgroesse wie folgt gelesen: Breite=%d,Hoehe=%d,Groesse=%d",BSlaveID,BFieldWidth,BFieldHeight,BFieldSize);
	  fflush(stderr);
      }
    BFieldBaseAddress   = malloc(BFieldSize);
    BFieldBufferAddress = malloc(BFieldSize);
    BFieldCurrentAddress= BFieldBaseAddress;

    if(BFieldBaseAddress == NULL)	/* Allokierung erfolgreich ? */
      {				/* nein */
	  fprintf (stderr, "Allokation von %d bytes fehlgeschlagen!\n",
		   BFieldSize);
	  if((DEBUG == ON) && (BDebugID == BSlaveID))
	    {
		fprintf(stderr,"\nPID[%3d] Fehler bei Allokierung des Speichers !",BSlaveID);
		fflush(stderr);
	    }
	  return(-1);
      }
    else
      {
	  if((DEBUG == ON) && (BDebugID == BSlaveID))
	    {
		fprintf(stderr,"\nPID[%3d] Speicher erfolgreich allokiert !",BSlaveID);
		fflush(stderr);
	    }
      }
    return(1);
}

void	BSetRules(int smin,int smax,int bmin,int bmax)
{
	BRuleSurviveMin	= smin;
	BRuleSurviveMax = smax;
	BRuleBirthMin = bmin;
	BRuleBirthMax = bmax;
}

int	BCalc(void)
{
	if((DEBUG == ON) && (BDebugID == BSlaveID))
	{
		fprintf(stderr,"\nPID[%3d] BCalc aufgerufen !",BSlaveID);
		fflush(stderr);
	}
	return(BsNewGeneration());
}

void	BSetDebugID(int id)
{
	BDebugID = id;
}

void	BSetSlaveID(int id)
{
	BSlaveID = id;
}

void	BEnd(void)
{
	free(BFieldBaseAddress);
	free(BFieldBufferAddress);
}

