#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include "PsxIO.h"
#include "comm.h"
#include "gui.h"
#include "master.h"

/*
 * Kanaele
 */

#define RING_LO0	4		/* links  aus Ring raus */
#define RING_LI0	5		/* links  in  Ring rein */
#define RING_LOI	7		/* rechts in  Ring rein */
#define RING_LII	6		/* rechts aus Ring raus */
#define PROC_IN(p)	(8+2*(p))	/* vom Slave p          */
#define PROC_OUT(p)	(9+2*(p))	/* zum Slave p          */

/*
 * global information
 */

int       procs;	/* Anzahl Prozessoren 			*/
clock_t   mtime;	/* Zeit seit Programmstart		*/
bmsize   *sizes;	/* Information ueber Aufgabe der Procs	*/
char     *empty_line;	/* leere Bitmap, eine Zeile gross	*/
char     *trash;	/* brauchen-wir-nicht-bitmap		*/

#define AKTTIME(s)	((double)(clock()-(s))/(double)CLK_TCK)

#define SEND(o,x,l,c) { \
			int xyz; \
			xyz = PsxWrite (o, (char *) (x), l); \
			if (xyz != l) { \
			  fprintf (stderr, "%8.2lf Fatal error master(%d): %d supposed %d errno=%d\n", \
				   AKTTIME(mtime), __LINE__, xyz, l, errno); \
			  c; \
			} \
		  };

#define RECV(i,x,l,c) { \
		        int xyz; \
			xyz = PsxRead (i, (char *) (x), l); \
			if (xyz != l) { \
			  fprintf (stderr, "%8.2lf Fatal error master(%d): %d supposed %d errno=%d\n", \
				   AKTTIME(mtime), __LINE__, xyz, l, errno); \
			  c; \
			} \
		  };

void RingUpdate (int lofl)
{
  SEND (RING_LI0,
	empty_line,
	lofl,
	exit (1));
  RECV (RING_LO0,
	trash,
	lofl,
	exit (1));
  RECV (RING_LII,
	trash,
	lofl,
	exit (1));
  SEND (RING_LOI,
	empty_line,
	lofl,
	exit (1));
}

void ProcExit (void)
{
    comm mycomm;

    mycomm.prozessor = 0;
    mycomm.command   = QUIT;

    SEND (RING_LI0, &mycomm, sizeof (comm), return);
    RECV (RING_LII, &mycomm, sizeof (comm), return);
}

void Calc (char *Bitmap, int xsize, int ysize, int steps,
		  int bmin, int bmax, int smin, int smax)
{
    clock_t tstep, ttotal=0;
    int index, thestep;
    comm mycomm;

    mycomm.prozessor = 0;
    mycomm.command   = COUNT;

    printf ("\nZaehle Prozesse ... "); fflush (stdout);
    SEND (RING_LI0, &mycomm, sizeof (comm), return);
    RECV (RING_LII, &mycomm, sizeof (comm), return);
    printf ("%d Prozesse gefunden.\n", procs=mycomm.prozessor);

    ShowBitmap  (Bitmap, xsize, ysize, 0, 0);
    FlushBitmap ();

    if ((sizes = (bmsize *) malloc (procs * sizeof (bmsize))) == NULL) {
	printf ("Hilfe! Ich kann mein Feld nicht anlegen!");
	return;
    }
    if ((empty_line = (char *) malloc ((xsize+7)/4)) == NULL) {
	printf ("Hilfe! Ich kann meine Bitmaps nicht anlegen!");
	return;
    }
    memset (empty_line, '\0', (xsize+7)/8);
    trash = empty_line + (xsize+7)/8;

    /*
     * Erzaehle den Prozessoren, wie gross deren Bitmap sein soll
     */
    mtime = clock ();

    for (index=0; index<procs; index++) {
	sizes[index].xsize  = xsize;
	sizes[index].ysize  = (ysize / procs) + ((index < (ysize%procs)) ? 1 : 0);
	sizes[index].yoffs  = index ? (sizes[index-1].yoffs + sizes[index-1].ysize) : 0;
	sizes[index].length = ((xsize + 7) / 8) * sizes[index].ysize;
	sizes[index].lofl   = ((xsize + 7) / 8) * 1;

	SEND (PROC_OUT(index), &(sizes[index]), sizeof (bmsize), exit (1));
    }

    /*
     * TODO: Lese und versende Bitmaps
     */
    for (index=0; index<procs; index++) {
	SEND (PROC_OUT(index),
	      Bitmap + sizes[index].yoffs * sizes[index].lofl,
	      sizes[index].length,
	      return);
    }

    /*
     * Update Ring
     */
    RingUpdate ((xsize+7)/8);

    /*
     * Send GO! command through ring
     */
    mycomm.command    = GO;
    mycomm.prozessor  = 0;
    mycomm.steps      = steps;
    mycomm.BirthMin   = bmin;
    mycomm.BirthMax   = bmax;
    mycomm.SurviveMin = smin;
    mycomm.SurviveMax = smax;
    
    SEND (RING_LI0, &mycomm, sizeof (comm), exit (1));
    RECV (RING_LII, &mycomm, sizeof (comm), exit (1));

    /*
     * Calc loop
     */
    for (thestep=0; thestep!=steps; thestep++) {
	clock_t ustart;
	tstep = clock ();
	/*
	 * Ring-Update links
	 */
	SEND (RING_LI0, empty_line, (xsize+7)/8, exit (1));
	RECV (RING_LO0, trash,      (xsize+7)/8, exit (1));

	/*
	 * Bitmaps lesen und auf den Bildschirm schicken.
	 */
	for (index=0; index<procs; index++) {
	    RECV (PROC_IN(index),
		  Bitmap + sizes[index].yoffs * sizes[index].lofl,
		  sizes[index].length,
		  exit (1));
	}

	/*
	 * Ringupdate rechts
	 */
	RECV (RING_LII, trash,      (xsize+7)/8, exit (1));
	SEND (RING_LOI, empty_line, (xsize+7)/8, exit (1));

	ustart = clock();
	ShowBitmap  (Bitmap, xsize, ysize, 0, 0);
	FlushBitmap ();

	printf ("%8.2lf Generation %3d: %8.2lf sec. U=%8.2lf\n",
		AKTTIME(mtime),
		thestep + 1,
		AKTTIME(tstep),
		AKTTIME(ustart));
	ttotal += clock () - tstep;

	Warte (Verzoegerung);
    }
    free (empty_line);
    free (sizes);
    return;
}
